import discord
import traceback
import sys

from discord import Interaction, app_commands
from discord.ext import commands
from typing import Optional, List

from disckit.utils import ErrorEmbed
from disckit.config import UtilConfig


class ErrorHandler(commands.Cog, name="Error Handler"):
    """Error handler for global application commands."""

    def __init__(self, bot: commands.Bot) -> None:
        self.bot = bot
        self.default_error_handler = app_commands.CommandTree.on_error

    async def cog_load(self) -> None:
        app_commands.CommandTree.on_error = self.on_error
        print(f"{self.__class__.__name__} has been loaded.")

    async def cog_unload(self) -> None:
        app_commands.CommandTree.on_error = self.default_error_handler
        print(f"{self.__class__.__name__} has been unloaded.")

    @staticmethod
    def __get_group_names(
        group: app_commands.Group,
        all_groups: Optional[List[app_commands.Group]] = None,
    ) -> List[str]:
        all_groups = all_groups or []
        all_groups.append(group.name)
        if group.parent is None:
            return all_groups
        ErrorHandler.__get_group_names(group.parent, all_groups)

    @staticmethod
    async def send_response(
        *,
        interaction: Interaction,
        embed: Optional[discord.Embed] = None,
        content: Optional[str] = None,
        ephemeral: bool = False,
    ) -> None:
        """Handles the error response to user."""

        load = {"ephemeral": ephemeral}
        if embed:
            load["embed"] = embed
        if content:
            load["content"] = content

        try:
            await interaction.response.send_message(**load)
        except discord.InteractionResponded:
            await interaction.followup.send(**load)

    @staticmethod
    async def throw_err(
        interaction: Interaction, error: discord.DiscordException
    ) -> None:
        print(
            f"Ignoring exception in command {interaction.command}:",
            file=sys.stderr,
        )
        traceback.print_exception(
            type(error), error, error.__traceback__, file=sys.stderr
        )

        channel = interaction.client.get_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        ) or await interaction.client.fetch_channel(
            UtilConfig.BUG_REPORT_CHANNEL
        )

        if channel is not None:
            if interaction.command:
                final_name = []
                if interaction.command.parent:
                    final_name = ErrorHandler.__get_group_names(
                        interaction.command.parent
                    )
                final_name.append(interaction.command.name)
                name = "/" + (" ".join(final_name))
            else:
                name = "Command not found"

            await channel.send(
                embed=ErrorEmbed(
                    f"```\nError caused by-\nAuthor Name: {interaction.user}"
                    f"\nAuthor ID: {interaction.user.id}\n"
                    f"\nError Type-\n{type(error)}\n"
                    f"\nError Type Description-\n{error.__traceback__.tb_frame}\n"
                    f"\nCause-\n{error.with_traceback(error.__traceback__)}```",
                    title=name,
                )
            )
        embed = ErrorEmbed(
            title="Sorry...",
            description="An unexpected error has occurred.\nThe developers have been notified of it.",
        )
        await ErrorHandler.send_response(interaction=interaction, embed=embed)

    async def on_error(
        self,
        interaction: discord.Interaction,
        error: app_commands.AppCommandError,
    ) -> None:
        error_embed = ErrorEmbed("Error")

        if isinstance(interaction.channel, discord.DMChannel):
            return

        elif (
            isinstance(error, commands.CommandError)
            and str(error) == "User is blacklisted."
        ):  # Custom error that is raised by disutils bots for blacklisting users.
            return

        elif isinstance(error, discord.NotFound):
            if error.code == 10008:
                return

        elif isinstance(error, commands.errors.NotOwner):
            error_embed.description = (
                "You do not have the required permissions to use this command.\n"
                "This command is only available to owners!"
            )
            await ErrorHandler.send_response(
                interaction=interaction, embed=error_embed
            )

        elif isinstance(error, app_commands.BotMissingPermissions):
            missing_permissions = ", ".join(error.missing_permissions)
            error_embed.description = (
                f"I don't have the required permissions for this command, "
                f"I need ``{missing_permissions}`` permission to proceed with this command."
            )
            error_embed.set_thumbnail(
                url="https://images.disutils.com/bot_assets/assets/missing_perms.png"
            )
            await ErrorHandler.send_response(
                interaction=interaction, embed=error_embed, ephemeral=True
            )

        elif isinstance(error, app_commands.MissingPermissions):
            missing_permissions = ", ".join(error.missing_permissions)
            error_embed.description = (
                f"You don't have the required permissions for this command, "
                f"you need ``{missing_permissions}`` permission to use this command."
            )
            error_embed.set_thumbnail(
                url="https://images.disutils.com/bot_assets/assets/access_denied.png"
            )
            await ErrorHandler.send_response(
                interaction=interaction, embed=error_embed, ephemeral=True
            )

        elif isinstance(error, app_commands.CommandSignatureMismatch):
            error_embed.description = (
                f"The signature of the command {error.command.name} seems to be different"
                " by the one provided by discord. To fix this issue please request the developers"
                " to sync the commands. If the issue still persists please contact the devs."
            )
            await ErrorHandler.send_response(
                interaction=interaction, embed=error_embed
            )

        else:
            await self.throw_err(interaction=interaction, error=error)


async def setup(bot: commands.Bot) -> None:
    await bot.add_cog(ErrorHandler(bot))



import discord

from discord.ext import commands, tasks
from typing import Iterator, Optional

from disckit.config import UtilConfig


class StatusHandler(commands.Cog, name="Status Handler"):
    """Cog for handling bot's dynamic status."""

    def __init__(self, bot: commands.Bot) -> None:
        self.bot = bot
        self.status: Optional[Iterator[str]] = None

    async def cog_load(self) -> None:
        self.status_task.start()
        print(f"{self.__class__.__name__} has been loaded.")

    async def cog_unload(self) -> None:
        self.status_task.cancel()
        print(f"{self.__class__.__name__} has been unloaded.")

    @tasks.loop(seconds=UtilConfig.STATUS_COOLDOWN)
    async def status_task(self) -> None:
        await self.bot.wait_until_ready()

        if self.status is None:
            self.status = await self.__get_iter()

        try:
            current_status = next(self.status)
        except StopIteration:
            self.status = await self.__get_iter()
            current_status = next(self.status)

        await self.bot.change_presence(
            activity=discord.Activity(
                type=UtilConfig.STATUS_TYPE, name=current_status
            )
        )

    async def __get_iter(self) -> Iterator[str]:
        return iter(
            await UtilConfig.STATUS_FUNC[0](
                self.bot, *UtilConfig.STATUS_FUNC[1]
            )
        )


async def setup(bot: commands.Bot) -> None:
    await bot.add_cog(StatusHandler(bot))



from __future__ import annotations

from discord import Colour, ActivityType, Locale
from typing import (
    Optional,
    TypeAlias,
    Union,
    ClassVar,
    Awaitable,
    Tuple,
    Dict,
    List,
    Set,
)
from enum import StrEnum
from disckit.utils import default_status_handler

TypeColor: TypeAlias = Union[int, Colour, Tuple[int, int, int]]
_BASE_WORKER_COG_PATH: str = "disckit.cogs.worker."


class UtilConfig:
    """The utility class which configures disckit's utilities."""

    MAIN_COLOR: ClassVar[TypeColor] = 0x5865F2
    """Class Attribute
    ---
    MAIN_COLOR: :class:`Optional[int, discord.color.Color, Tuple[int, int, int]]`

        The color of the MainEmbed.
    """

    SUCCESS_COLOR: ClassVar[TypeColor] = 0x00FF00
    """Class Attribute
    ---
    SUCCESS_COLOR: :class:`Optional[int, discord.color.Color, Tuple[int, int, int]]`

        The color of the SuccessEmbed.
    """

    ERROR_COLOR: ClassVar[TypeColor] = 0xFF0000
    """Class Attribute
    ---
    ERROR_COLOR: :class:`Optional[int, discord.color.Color, Tuple[int, int, int]]`

        The color of the ErrorEmbed.
    """

    SUCCESS_EMOJI: ClassVar[str] = "✅"
    """Class Attribute
    ---
    SUCCESS_EMOJI: :class:`Optional[str]`

        An emoji used in the title of the SuccessEmbed.
    """

    ERROR_EMOJI: ClassVar[str] = "❌"
    """Class Attribute
    ---
    ERROR_EMOJI: :class:`Optional[str]`

        An emoji used in the title of the ErrorEmbed.
    """

    FOOTER_IMAGE: ClassVar[Optional[str]] = None
    """Class Attribute
    ---
    FOOTER_IMAGE: :class:`Optional[str]`

        A URL to an image for the footer of `MainEmbed`, `SuccessEmbed` and `ErrorEmbed`.
    """

    FOOTER_TEXT: ClassVar[Optional[str]] = None
    """Class Attribute
    ---
    FOOTER_TEXT: :class:`Optional[str]`

        The footer text of `MainEmbed`, `SuccessEmbed` and `ErrorEmbed`.
    """

    STATUS_FUNC: ClassVar[
        Tuple[Awaitable[Union[Tuple[str, ...], List[str], Set[str]]], Tuple]
    ] = (
        default_status_handler,
        (),
    )
    """Class Attribute
    ---
    STATUS_FUNC: :class:`Tuple[Awaitable[Union[Tuple, List, Set]], Tuple]`

        A tuple having its first element as a coroutine object which will be awaited when-
        
        - When the cog first loads.
        - When the handler is done iterating through all statuses returned from the function.

        The second element is a tuple containing the extra arguments that can be passed to your
        custom status handler function. If no arguments have to be passed an empty tuple
        should suffice.
    """

    STATUS_TYPE: ClassVar[ActivityType] = ActivityType.listening
    """Class Attribute
    ---
    STATUS_TYPE: :class:`ActivityType`

        The discord acitvity type used by the StatusHandler.
    """

    STATUS_COOLDOWN: ClassVar[Optional[int]] = None
    """Class Attribute
    ---
    STATUS_COOLDOWN: :class:`Optional[int]`

        A cooldown in seconds for how long a status will play before changing in the `StatusHandler` cog.
    """

    BUG_REPORT_CHANNEL: ClassVar[Optional[int]] = None
    """Class Attribute
    ---
    BUG_REPORT_CHANNEL: :class:`Optional[int]`

        The channel ID to where the bug reports will be sent to by the `ErrorHandler` cog.
    """

    LEMMA_TRANS_GROUPS: ClassVar[Optional[Dict[Locale, str]]] = None
    """Class Attribute
    ---

    LEMMA_TRANS_PATHS: :class:`Optional[Dict[Locale, str]]`

        A :class:`discord.Locale` object mapping to the folder / file containing it's respective group
        translation files only. The LemmaTranslator only works for translating command paramters and
        description only.
    """

    LEMMA_TRANS_COMMANDS: ClassVar[Optional[Dict[Locale, str]]] = None
    """Class Attribute
    ---

    LEMMA_TRANS_PATHS: :class:`Optional[Dict[Locale, str]]`

        A :class:`discord.Locale` object mapping to the folder / file containing it's respective command
        translation files. The LemmaTranslator only works for translating command paramters and description
        only.
    """


class CogEnum(StrEnum):
    ERROR_HANDLER: str = _BASE_WORKER_COG_PATH + "error_handler"
    """An extension for error handling."""

    STATUS_HANDLER: str = _BASE_WORKER_COG_PATH + "status_handler"
    """An extension for the bot's status handling."""



from discord.ext import commands
from typing import Optional

from disckit.config import CogEnum, UtilConfig
from disckit.errors import CogLoadError


async def dis_load_extension(
    bot: commands.Bot,
    *cogs: CogEnum,
    debug_message: Optional[str] = "Loading extension: {}",
) -> None:
    """|coro|
    A custom extension loader specifically for the disckit cogs.

    Parameters
    ----------
    bot: :class:`commands.Bot`
        The bot instance.

    *cogs: :class:`CogEnum`
        The cogs to be loaded from disckit package.

    debug_message: :class:`Optional[str]`
        The debug message to be printed out when the cog is loaded.
        Needs to contain one `{}` which is formatted to the cog name
        being loaded.

    Raises
    ------
    :class:`CogLoadError`
        Raised when an error occurrs in loading the cog.
    """

    message = None
    for cog in set(cogs):
        if cog == CogEnum.STATUS_HANDLER:
            if not UtilConfig.STATUS_FUNC:
                message = (
                    "Attribute - `UtilConfig.STATUS_FUNC` needs"
                    " to be set to use StatusHandler cog"
                )

            elif not UtilConfig.STATUS_TYPE:
                message = (
                    "Attribute - `UtilConfig.STATUS_TYPE` needs"
                    "to be set to use StatusHandler cog"
                )

            elif not UtilConfig.STATUS_COOLDOWN:
                message = (
                    "Attribute - `UtilConfig.STATUS_COOLDOWN` needs"
                    " to be set to use StatusHandler cog."
                )

        if cog == CogEnum.ERROR_HANDLER:
            if not UtilConfig.BUG_REPORT_CHANNEL:
                message = (
                    "Attribute - `UtilConfig.BUG_REPORT_CHANNEL` needs"
                    " to be set to use ErrorHandler cog."
                )

        if message:
            raise CogLoadError(message=message, cog=cog)

        await bot.load_extension(cog.value)
        if debug_message:
            print(debug_message.format(cog.name.title().replace(" ", "")))



from discord import Embed, utils
from typing import Optional, TYPE_CHECKING

import disckit

if TYPE_CHECKING:
    import disckit.config


class MainEmbed(Embed):
    """Represents a main embed for general use."""

    def __init__(
        self, description: Optional[str] = None, title: Optional[str] = None
    ) -> None:
        """
        Parameters
        ----------
        description: :class:`str`
            The description of the main embed.

        title: :class:`str`, default `None`
            The title of the main embed.
        """

        super().__init__(
            title=title,
            description=description,
            color=disckit.config.UtilConfig.MAIN_COLOR,
            timestamp=utils.utcnow(),
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )


class SuccessEmbed(Embed):
    """Represents a success embed."""

    def __init__(
        self, description: Optional[str] = None, title: Optional[str] = None
    ) -> None:
        """
        Parameters
        ----------
        description: :class:`str`
            The description of the success embed.

        title: :class:`str`, default `None`
            The title of the success embed.
        """

        if title:
            title = f"{disckit.config.UtilConfig.SUCCESS_EMOJI} {title}"
        super().__init__(
            title=title,
            description=description,
            color=disckit.config.UtilConfig.SUCCESS_COLOR,
            timestamp=utils.utcnow(),
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )


class ErrorEmbed(Embed):
    """Represents an error embed."""

    def __init__(
        self, description: Optional[str] = None, title: Optional[str] = None
    ) -> None:
        """
        Parameters
        ----------
        description: :class:`str`
            The description of the error embed.

        title: :class:`str`, default `None`
            The title of the error embed.
        """

        if title:
            title = f"❌ {title}"
        super().__init__(
            title=title,
            description=description,
            color=disckit.config.UtilConfig.ERROR_COLOR,
            timestamp=utils.utcnow(),
        )
        self.set_footer(
            text=disckit.config.UtilConfig.FOOTER_TEXT,
            icon_url=disckit.config.UtilConfig.FOOTER_IMAGE,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from disckit.config import CogEnum


class DisException(Exception):
    """Base class of disckit's exceptions."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CogLoadError(DisException):
    """Raised when loading a cog fails.

    Attributes
    ------------
    cog: :class:`CogEnum`
        The cog that failed loading.
    """

    def __init__(self, message: str, cog: CogEnum, **kwargs: Any) -> None:
        super().__init__(message, **kwargs)
        self.cog = cog


class LemmaLoadError(DisException):
    """Raised when an error occurs in loading the translator

    Attributes
    ------------
    error_code: :class:`int`
        The error code of the exception.
    """

    def __init__(self, message: str, error_code: int, **kwargs: Any) -> None:
        super().__init__(message, **kwargs)
        self.error_code = error_code



from __future__ import annotations

import discord
import functools

from typing import TYPE_CHECKING
from discord.app_commands import Choice
from disckit.utils.embeds import SuccessEmbed, MainEmbed, ErrorEmbed

if TYPE_CHECKING:
    from discord import Interaction
    from discord.ext.commands import Bot
    from typing import (
        Any,
        Awaitable,
        Callable,
        Coroutine,
        List,
        Tuple,
        TypeVar,
    )

    _T = TypeVar("_T", str, int, float)


__all__ = (
    "MainEmbed",
    "SuccessEmbed",
    "ErrorEmbed",
    "default_status_handler",
    "make_autocomplete",
)


async def default_status_handler(bot: Bot, *args: Any) -> Tuple[str, ...]:
    """The default status handler. The first parameter will always be the
    bot instance which will automatically be passed as argument in the
    status handler.

    This function is called when cog first loads and when the handler is
    done iterating through all the statuses returned from the function.


    Parameters
    ----------
    bot: :class:`commands.Bot`
        The global bot instance that gets passed to the function automatically.

    *args: :class:`Any`
        The extra arguments passed in `UtilUtilConfig.STATUS_FUNC[1]`
        (The second element is the extra arguments that will be passed on).

    Returns
    --------
    :class:`Tuple` [:class:`str`, ...]
        Heehee hawhaw
    """

    users = len(bot.users)
    guilds = len(bot.guilds)
    status = (
        # Prefixed by "Listening to" as the default ActivityType
        # (UtilConfig.STATUS_TYPE = ActivityType.listening).
        f"{users:,} users",
        f"humans from {guilds:,} servers",
        "Slash commands!",
    )

    return status


def make_autocomplete(
    *args: _T,
) -> Callable[[Interaction, str], Awaitable[list[Choice[_T]]]]:
    """
    Creates an autocomplete function for the given arguments.

    Parameters
    ----------
        *args: :class:`str`: Options for the autocomplete

    Returns
    --------
        A function that can be put in @discord.app_commands.autocomplete

    Usage
    ------
        ```
        @app_commands.autocomplete(choice=make_autocomplete("Heads", "Tails"))
        @app_commands.command(name="coin-flip")
        async def coin_flip(
            self, interaction: discord.Interaction, choice: str
        ): ...
        ```
    """
    choices = [Choice(name=str(arg), value=arg) for arg in args]

    async def autocomplete(_, __) -> List[Choice[_T]]:  # noqa ANN001
        return choices

    return autocomplete


async def sku_check(bot: Bot, sku_id: int, user_id: int) -> bool:
    """|coro|

    Checks if a user has purchased a specific SKU package.

    Parameters
    ----------
    bot: :class:`int`
        The bot class.
    sku_id: :class:`int`
        The SKU ID of the package.
    user_id: :class:`int`
        The Discord user ID to check.
    """

    sku = discord.Object(id=sku_id)
    user = discord.Object(id=user_id)

    user_entitlements = [
        entitlement
        async for entitlement in bot.entitlements(skus=[sku], user=user)
    ]

    return bool(user_entitlements)


def disallow_bots(
    func: Coroutine[Any, Any, _T],
) -> None | Coroutine[Any, Any, _T]:
    """A decorator used for not allowing members to pass in a bot user into command params"""

    @functools.wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> None:
        interaction: None | Interaction[Bot] = None
        bot_user: bool = False

        for arg in args + tuple(kwargs.values()):
            if isinstance(arg, discord.Interaction):
                interaction = arg

            elif isinstance(arg, (discord.Member, discord.User)):
                bot_user = bot_user or arg.bot

        if bot_user and interaction:
            embed = ErrorEmbed("You cannot interact with bots!")
            try:
                await interaction.response.send_message(
                    embed=embed, ephemeral=True
                )
            except discord.InteractionResponded:
                await interaction.followup.send(embed=embed, ephemeral=True)
            return
        await func(*args, **kwargs)

    return wrapper


def is_owner(
    func: Coroutine[Any, Any, _T],
) -> None | Coroutine[Any, Any, _T]:
    """A decorator for owner-only slash commands"""

    @functools.wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> None:
        interaction: None | Interaction[Bot] = None

        for arg in args + tuple(kwargs.values()):
            if isinstance(arg, discord.Interaction):
                interaction = arg
                break

        if interaction and interaction.user.id in interaction.client.owner_ids:
            await func(*args, **kwargs)
        else:
            embed = ErrorEmbed("This command is owner only!")
            try:
                await interaction.response.send_message(
                    embed=embed, ephemeral=True
                )
            except discord.InteractionResponded:
                await interaction.followup.send(embed=embed, ephemeral=True)
            return

    return wrapper



import re
from setuptools import setup


def derive_version() -> str:
    version = ""
    with open("src/disckit/__init__.py") as f:
        version = re.search(
            r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
        ).group(1)

    if not version:
        raise RuntimeError("Version is not set.")

    return version


setup(version=derive_version())



"""
Disutils' utility package
~~~~~~~~~~~~~~~~~~~~~~~~~

A utility package made for the disutils bots.

:copyright: (c) 2024-present Disutils Team
:license: MIT, see LICENSE for more details.

"""

__version__ = "0.7"
__title__ = "disckit"
__author__ = "Jiggly Balls"
__license__ = "MIT"
__copyright__ = "Copyright 2024-present Disutils Team"

from disckit.config import UtilConfig, CogEnum
from disckit.errors import DisException, CogLoadError
from typing import NamedTuple, Literal

__all__ = ("UtilConfig", "CogEnum", "DisException", "CogLoadError")


class VersionInfo(NamedTuple):
    major: int
    minor: int
    release_level: Literal["alpha", "beta", "final"]


def _expand() -> VersionInfo:
    v = __version__.split(".")
    level_types = {"a": "alpha", "b": "beta"}
    level = level_types.get(v[-1], "final")
    return VersionInfo(major=v[0], minor=v[1], release_level=level)


version_info: VersionInfo = _expand()
